"""PDBx/mmCIF Python dictionary resources.

All of the code in this repository is original based on
http://mmcif.wwpdb.org/. Specifically, this code is directly derived from the
`pdbx code <http://mmcif.wwpdb.org/docs/sw-examples/python/src/pdbx.tar.gz>`_
linked from
`PDBx Python Parser Examples and Tutorial
<http://mmcif.wwpdb.org/docs/sw-examples/python/html/>`_.

See `PDBx Python Parser Examples and Tutorial
<http://mmcif.wwpdb.org/docs/sw-examples/python/html/>`_ for more information
about this package, including examples.
"""

# import pdbx.reader
import io
from .reader import PdbxReader
from .writer import PdbxWriter
from .errors import PdbxSyntaxError, PdbxError  # noqa: F401
from .containers import DataCategory, DataContainer  # noqa: F401
from ._version import __version__  # noqa: F401


def load(fobj) -> list:
    """Parse a CIF file.

    :param file fobj:  file object ready for reading
    :returns:  a list of :class:`~pdbx.containers.DataContainer` objects
    """
    data = []
    PdbxReader(fobj).read(data)
    return data


def loads(text) -> list:
    """Parse a CIF string.

    :param str s:  string with CIF data
    :returns: a list of :class:`~pdbx.containers.DataContainer` objects
    """
    return load(io.StringIO(text))


def dump(datacontainers, fobj):
    """Write a list of objects to a CIF file.

    :param list datacontainers:  a list of :class:`~pdbx.containers.DataContainer` objects # noqa E501
    :param file fobj:  a file object ready for writing
    """
    PdbxWriter(fobj).write(datacontainers)


def dumps(datacontainers) -> str:
    """Serialize a list of objects to a CIF-formatted string.

    :param list datacontainers:  list of :class:`~pdbx.containers.DataContainer` objects # noqa E501
    :returns:  CIF-formatted string
    """
    fobj = io.StringIO()
    dump(datacontainers, fobj)
    return fobj.getvalue()



"""Error classes for PDBx/mmCIF."""


class PdbxError(Exception):
    """Class for general errors."""


class PdbxSyntaxError(Exception):
    """Class for syntax errors."""

    def __init__(self, line_number, text):
        super().__init__(self)
        self.line_number = line_number
        self.text = text

    def __str__(self):
        return "%%ERROR - [at line: %d] %s" % (self.line_number, self.text)



##
# File: PdbxWriter.py
# Date: 2011-10-09 Jdw Adapted from PdbxParser.py
#
# Updates:
# 5-Apr-2011 jdw Using the double quote format preference
# 23-Oct-2012 jdw update path details and reorganize.
#
###
"""Classes for writing data and dictionary containers in PDBx/mmCIF format."""
from sys import stdout
from .containers import DefinitionContainer, DataContainer
from .errors import PdbxError

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"


MAXIMUM_LINE_LENGTH = 2048
SPACING = 2
INDENT_DEFINITION = 3
DO_DEFINITION_INDENT = False


class PdbxWriter:
    """Write PDBx data files or dictionaries.
    Use the input container or container list.
    """

    def __init__(self, output_file=stdout):
        """Initialize.

        :param file output_file:  file object ready for writing
        """
        self.__output_file = output_file
        self.__container_list = []
        self.__maximum_line_length = MAXIMUM_LINE_LENGTH
        self.__spacing = SPACING
        self.__indent_definition = INDENT_DEFINITION
        self.__indent_space = " " * self.__indent_definition
        self._do_definition_indent = DO_DEFINITION_INDENT
        # Maximum number of rows checked for value length and format
        self.__row_partition = None

    def set_row_partition(self, num_rows):
        """Maximum number of rows checked for value length and format.

        :param int num_rows:  maximum number of rows
        """
        self.__row_partition = num_rows

    def write(self, container_list):
        """Write out a list of containers.

        :param list container_list:  list of
          :class:`~pdbx.containers.ContainerBase` objects to write.
        """
        self.__container_list = container_list
        for container in self.__container_list:
            self.write_container(container)

    def write_container(self, container):
        """Write out information for an individual container.

        :param container:  container to write
        :type container:  :class:`~pdbx.containers.ContainerBase`
        """
        indent_string = " " * self.__indent_definition
        if isinstance(container, DefinitionContainer):
            self.__write("save_%s\n" % container.name)
            self._do_definition_indent = True
            self.__write(indent_string + "#\n")
        elif isinstance(container, DataContainer):
            if container.get_global():
                self.__write("global_\n")
                self._do_definition_indent = False
                self.__write("\n")
            else:
                self.__write("data_%s\n" % container.name)
                self._do_definition_indent = False
                self.__write("#\n")
        for name in container.get_object_name_list():
            obj = container.get_object(name)
            object_list = obj.row_list
            # Skip empty objects
            if not object_list:
                continue
            # Item - value formattting
            if len(object_list) == 1:
                self.__write_item_value_format(obj)
            # Table formatting
            elif len(object_list) > 1 and obj.attribute_list:
                self.__write_table_format(obj)
            else:
                raise PdbxError(
                    "len(object_list) = %d and len(obj.attribute_list) = %d"
                    % (len(object_list), len(obj.attribute_list))
                )
            if self._do_definition_indent:
                self.__write(indent_string + "#")
            else:
                self.__write("#")
        # Add a trailing saveframe reserved word
        if isinstance(container, DefinitionContainer):
            self.__write("\nsave_\n")
        self.__write("#\n")

    def __write(self, string_):
        """Write a string.

        :param str string_:  string to write
        """
        self.__output_file.write(string_)

    def __write_item_value_format(self, category):
        """Write items and values for the given category.

        :param category:  category to write
        :type category:  :class:`~pdbx.containers.DataCategory`
        """
        # Compute the maximum item name length within this category -
        attribute_name_max_length = 0
        for attribute_name in category.attribute_list:
            attribute_name_max_length = max(
                attribute_name_max_length, len(attribute_name)
            )
        item_name_max_length = (
            self.__spacing + len(category.name) + attribute_name_max_length + 2
        )
        line_list = []
        line_list.append("#\n")
        for attribute_name, _ in category.attribute_list_with_order:
            if self._do_definition_indent:
                # - add indent --
                line_list.append(self.__indent_space)
            item_name = "_%s.%s" % (category.name, attribute_name)
            line_list.append(item_name.ljust(item_name_max_length))
            line_list.append(category.get_value_formatted(attribute_name, 0))
            line_list.append("\n")
        self.__write("".join(line_list))

    def __write_table_format(self, category):
        """Write table format data.

        :param category:  category to write
        :type category:  :class:`~pdbx.containers.DataCategory`
        """
        # Write the declaration of the loop_
        line_list = []
        line_list.append("#\n")
        if self._do_definition_indent:
            line_list.append(self.__indent_space)
        line_list.append("loop_")
        for attribute_name in category.attribute_list:
            line_list.append("\n")
            if self._do_definition_indent:
                line_list.append(self.__indent_space)
            item_name = "_%s.%s" % (category.name, attribute_name)
            line_list.append(item_name)
        self.__write("".join(line_list))
        # Write the data in tabular format
        # For speed make the following evaluation on a portion of the table
        if self.__row_partition is not None:
            num_steps = max(1, category.row_count / self.__row_partition)
        else:
            num_steps = 1
        format_type_list, _ = category.get_format_type_list(steps=num_steps)
        max_length_list = category.get_max_attribute_list_length(
            steps=num_steps
        )
        spacing = " " * self.__spacing
        for irow in range(category.row_count):
            line_list = []
            line_list.append("\n")
            if self._do_definition_indent:
                line_list.append(self.__indent_space + " ")
            for iattr in range(category.attribute_count):
                format_type = format_type_list[iattr]
                max_length = max_length_list[iattr]
                if format_type in ("FT_UNQUOTED_STRING", "FT_NULL_VALUE"):
                    val = category.get_value_formatted_by_index(iattr, irow)
                    line_list.append(val.ljust(max_length))
                elif format_type == "FT_NUMBER":
                    val = category.get_value_formatted_by_index(iattr, irow)
                    line_list.append(val.rjust(max_length))
                elif format_type == "FT_QUOTED_STRING":
                    val = category.get_value_formatted_by_index(iattr, irow)
                    line_list.append(val.ljust(max_length + 2))
                elif format_type == "FT_MULTI_LINE_STRING":
                    val = category.get_value_formatted_by_index(iattr, irow)
                    line_list.append(val)
                line_list.append(spacing)
            self.__write("".join(line_list))
        self.__write("\n")



##
# File: PdbxReader.py
# Date: 2012-01-09 Jdw Adapted from PdbxParser
#
# Updates:
#
# 2012-01-09 - (jdw) Separate reader and writer classes.
#
# 2012-09-02 - (jdw) Revise tokenizer to better handle embedded quoting.
#
##
"""PDBx/mmCIF dictionary and data file parser.

.. note::

   Acknowledgements:

   The tokenizer used in this module is modeled after the clever parser
   design used in the PyMMLIB package.

   PyMMLib Development Group:

   Authors: Ethan Merritt: merritt@u.washington.edu, Jay Painter: jay.painter@gmail.com

   See: http://pymmlib.sourceforge.net/
"""
import re
from .containers import DataCategory, DefinitionContainer, DataContainer
from .errors import PdbxSyntaxError


class PdbxReader:
    """PDBx reader for data files and dictionaries."""

    def __init__(self, input_file):
        """Initialize.

        :param file input_file: input file handle; e.g. as returned by open().
        """
        self.__current_line_number = 0
        self.__input_file = input_file
        self.__state_dict = {
            "data": "ST_DATA_CONTAINER",
            "loop": "ST_TABLE",
            "global": "ST_GLOBAL_CONTAINER",
            "save": "ST_DEFINITION",
            "stop": "ST_STOP",
        }

    def read(self, container_list):
        """Appends to the input list of definition and data containers.

        :param list container_list:  list of :class:`~pdbx.containers.ContainerBase` containers to append to.
        """
        self.__current_line_number = 0
        try:
            self.__parser(self.__tokenizer(self.__input_file), container_list)
        except StopIteration:
            self.__syntax_error("Unexpected end of file")

    def __syntax_error(self, error_text):
        """Raise a PdbxSyntaxError.

        :param str error_text:  text for exception message
        :raises pdbx.errors.PdbxSyntaxError:  exception with error text
        """
        raise PdbxSyntaxError(self.__current_line_number, error_text)

    @staticmethod
    def __get_container_name(in_word) -> str:
        """Returns the name of the data_ or save_ container.

        :param str in_word:  input word
        """
        return str(in_word[5:]).strip()

    def __get_state(self, in_word) -> tuple:
        """Identifies reserved syntax elements and assigns an associated state.

        :param str in_word:  input word
        :returns: (reserved word, state) where:

          * reserved word - is one of CIF syntax elements: data_, loop_,
          global_, save_, stop_

          * state - the parser state required to process this next section.
        """
        i = in_word.find("_")
        if i == -1:
            return None, "ST_UNKNOWN"
        try:
            reserved_word = in_word[:i].lower()
            return reserved_word, self.__state_dict[reserved_word]
        except KeyError:
            return None, "ST_UNKNOWN"

    def __parser(self, tokenizer, container_list):
        """Parser for PDBx data files and dictionaries.

        :param tokenizer: reentrant method recognizing data item names
          (_category.attribute), quoted strings (single, double and
          multi-line semi-colon delimited), and unquoted strings.
        :param list container_list: list-type container for data and
          definition objects parsed from from the input file.
          container_list is appended with data and definition objects.
        """
        # Working container - data or definition
        current_container = None
        # Working category container
        category_index = {}
        current_category = None
        current_row = None
        state = None

        # Find the first reserved word and begin capturing data.
        for (
            current_category_name,
            current_attribute_name,
            current_quoted_string,
            current_word,
        ) in tokenizer:
            if current_word is None:
                continue
            reserved_word, state = self.__get_state(current_word)
            if reserved_word is not None:
                break
        else:
            # empty file
            return

        while True:
            # Set the current state: at this point in the processing cycle we
            # are expecting a token containing # either a '_category.attribute'
            # or a reserved word.
            if current_category_name is not None:
                state = "ST_KEY_VALUE_PAIR"
            elif current_word is not None:
                reserved_word, state = self.__get_state(current_word)
            else:
                self.__syntax_error("Miscellaneous syntax error")
                return

            # Process _category.attribute value assignments
            if state == "ST_KEY_VALUE_PAIR":
                try:
                    current_category = category_index[current_category_name]
                except KeyError:
                    # A new category is encountered - create a container and
                    # add a row
                    category_index[current_category_name] = DataCategory(
                        current_category_name
                    )
                    current_category = category_index[current_category_name]
                    try:
                        current_container.append(current_category)
                    except AttributeError:
                        self.__syntax_error(
                            "Category cannot be added to data_ block"
                        )
                        return
                    current_row = []
                    current_category.append(current_row)
                else:
                    # Recover the existing row from the category
                    try:
                        current_row = current_category[0]
                    except IndexError:
                        self.__syntax_error(
                            "Internal index error accessing category data"
                        )
                        return
                # Check for duplicate attributes and add attribute to table.
                if current_attribute_name in current_category.attribute_list:
                    self.__syntax_error(
                        "Duplicate attribute encountered in category"
                    )
                    return
                else:
                    current_category.append_attribute(current_attribute_name)
                # Get the data for this attribute from the next token
                tok_category, _, current_quoted_string, current_word = next(
                    tokenizer
                )
                if tok_category is not None or (
                    current_quoted_string is None and current_word is None
                ):
                    self.__syntax_error(
                        "Missing data for item _%s.%s"
                        % (current_category_name, current_attribute_name)
                    )
                if current_word == "?":
                    current_row.append(None)
                elif current_word == ".":
                    current_row.append("")
                elif current_word is not None:
                    # Validation check token for misplaced reserved words
                    reserved_word, state = self.__get_state(current_word)
                    if reserved_word is not None:
                        self.__syntax_error(
                            "Unexpected reserved word: %s" % (reserved_word)
                        )
                    current_row.append(current_word)
                elif current_quoted_string is not None:
                    current_row.append(current_quoted_string)
                else:
                    self.__syntax_error("Missing value in item-value pair")
                try:
                    (
                        current_category_name,
                        current_attribute_name,
                        current_quoted_string,
                        current_word,
                    ) = next(tokenizer)
                except StopIteration:
                    return
                continue

            # Process a loop_ declaration and associated data
            if state == "ST_TABLE":
                # The category name in the next current_category_name,
                # current_attribute_name pair defines the name of the category
                # container.
                (
                    current_category_name,
                    current_attribute_name,
                    current_quoted_string,
                    current_word,
                ) = next(tokenizer)
                if (current_category_name is None) or (
                    current_attribute_name is None
                ):
                    self.__syntax_error(
                        "Unexpected token in loop_ declaration"
                    )
                    return
                # Check for a previous category declaration.
                if current_category_name in category_index:
                    self.__syntax_error(
                        "Duplicate category declaration in loop_"
                    )
                    return
                current_category = DataCategory(current_category_name)
                try:
                    current_container.append(current_category)
                except AttributeError:
                    self.__syntax_error(
                        "loop_ declaration outside of data_ block or save_ "
                        "frame"
                    )
                    return
                current_category.append_attribute(current_attribute_name)
                # Read the rest of the loop_ declaration
                for (
                    current_category_name,
                    current_attribute_name,
                    current_quoted_string,
                    current_word,
                ) in tokenizer:
                    if current_category_name is None:
                        break
                    if current_category_name != current_category.name:
                        self.__syntax_error(
                            "Changed category name in loop_ declaration"
                        )
                        return
                    current_category.append_attribute(current_attribute_name)
                else:
                    # formal CIF 1.1 grammar expects at least one value
                    self.__syntax_error("loop_ without values")
                # If the next token is a 'word', check it for any reserved
                # words
                if current_word is not None:
                    reserved_word, state = self.__get_state(current_word)
                    if reserved_word is not None:
                        if reserved_word == "stop":
                            return
                        else:
                            self.__syntax_error(
                                "Unexpected reserved word after loop "
                                "declaration: %s" % (reserved_word)
                            )
                # Read the table of data for this loop_
                while True:
                    current_row = []
                    current_category.append(current_row)
                    for _ in current_category.attribute_list:
                        if current_word == "?":
                            current_row.append(None)
                        elif current_word == ".":
                            current_row.append("")
                        elif current_word is not None:
                            current_row.append(current_word)
                        elif current_quoted_string is not None:
                            current_row.append(current_quoted_string)
                        try:
                            (
                                current_category_name,
                                current_attribute_name,
                                current_quoted_string,
                                current_word,
                            ) = next(tokenizer)
                        except StopIteration:
                            return
                    # loop_ data processing ends if a new _category.attribute
                    # is encountered
                    if current_category_name is not None:
                        break
                    # A reserved word is encountered
                    if current_word is not None:
                        reserved_word, state = self.__get_state(current_word)
                        if reserved_word is not None:
                            break
                continue

            if state == "ST_DEFINITION":
                # Ignore trailing unnamed saveframe delimiters e.g. 'save_'
                state_name = self.__get_container_name(current_word)
                if state_name:
                    current_container = DefinitionContainer(state_name)
                    container_list.append(current_container)
                    category_index = {}
                    current_category = None
            elif state == "ST_DATA_CONTAINER":
                data_name = self.__get_container_name(current_word)
                if not data_name:
                    data_name = "unidentified"
                current_container = DataContainer(data_name)
                container_list.append(current_container)
                category_index = {}
                current_category = None
            elif state == "ST_STOP":
                return
            elif state == "ST_GLOBAL":
                current_container = DataContainer("blank-global")
                current_container.set_global()
                container_list.append(current_container)
                category_index = {}
                current_category = None
            elif state == "ST_UNKNOWN":
                self.__syntax_error(
                    "Unrecognized syntax element: " + str(current_word)
                )
                return
            else:
                assert False, f"unhandled state {state}"

            try:
                (
                    current_category_name,
                    current_attribute_name,
                    current_quoted_string,
                    current_word,
                ) = next(tokenizer)
            except StopIteration:
                return

    def __tokenizer(self, input_file):
        """Tokenizer method for the mmCIF syntax file.

        Each return/yield from this method returns information about the next
        token in the form of a tuple with the following structure:
            (category name, attribute name, quoted strings, words w/o quotes
            or white space)
        Differentiated the regular expression to the better handle embedded
        quotes.

        :param file input_file:  file object ready for reading
        :rtype: Iterator[tuple]
        """
        # Regex definition for mmCIF syntax - semi-colon delimited strings are
        # handled outside of this regex.
        mmcif_re = re.compile(
            r"(?:"
            r"(?:_(.+?)[.](\S+))"
            "|"  # _category.attribute
            r"(?:['](.*?)(?:[']\s|[']$))"
            "|"  # single quoted strings
            r"(?:[\"](.*?)(?:[\"]\s|[\"]$))"
            "|"  # double quoted strings
            r"(?:\s*#.*$)"
            "|"  # comments (dumped)
            r"(\S+)"  # unquoted words
            r")"
        )
        file_iterator = iter(input_file)
        # Tokenizer loop begins here
        for line in file_iterator:
            self.__current_line_number += 1
            # Dump comments
            if line.startswith("#"):
                continue
            # Gobble up the entire semi-colon/multi-line-delimited string and
            # and stuff this into the string slot in the return tuple
            if line.startswith(";"):
                multiline_string = [line[1:]]
                for line in file_iterator:
                    self.__current_line_number += 1
                    if line.startswith(";"):
                        break
                    multiline_string.append(line)
                else:
                    self.__syntax_error("unterminated multi-line string")
                # remove trailing new-line that is part of the \n; delimiter
                multiline_string[-1] = multiline_string[-1].rstrip()
                yield (None, None, "".join(multiline_string), None)
                # Need to process the remainder of the current line -
                line = line[1:]

            # Apply regex to the current line consolidate the single/double
            # quoted within the quoted string category
            for match in mmcif_re.finditer(line):
                match_groups = match.groups()
                if match_groups != (None, None, None, None, None):
                    if match_groups[2] is not None:
                        quoted_string = match_groups[2]
                    elif match_groups[3] is not None:
                        quoted_string = match_groups[3]
                    else:
                        quoted_string = None
                    groups = (
                        match_groups[0],
                        match_groups[1],
                        quoted_string,
                        match_groups[4],
                    )
                    yield groups

    def __tokenizer_org(self, input_file):
        """Tokenizer method for the mmCIF syntax file.

        Each return/yield from this method returns information about the next
        token in the form of a tuple with the following structure:
            (category name, attribute name, quoted strings, words w/o quotes
            or white space)

        :param file input_file:  file object ready for reading
        :rtype: Iterator[tuple]
        """
        # Regex definition for mmCIF syntax - semi-colon delimited strings are
        # handled outside of this regex.
        mmcif_re = re.compile(
            r"(?:"
            r"(?:_(.+?)[.](\S+))"
            "|"  # _category.attribute
            r"(?:['\"](.*?)(?:['\"]\s|['\"]$))"
            "|"  # quoted strings
            r"(?:\s*#.*$)"
            "|"  # comments (dumped)
            r"(\S+)"  # unquoted words
            r")"
        )
        file_iterator = iter(input_file)
        # Tokenizer loop begins here
        while True:
            line = next(file_iterator)
            self.__current_line_number += 1
            # Dump comments
            if line.startswith("#"):
                continue
            # Gobble up the entire semi-colon/multi-line delimited string and
            # and stuff this into the string slot in the return tuple
            if line.startswith(";"):
                multiline_string = [line[1:]]
                while True:
                    line = next(file_iterator)
                    self.__current_line_number += 1
                    if line.startswith(";"):
                        break
                    multiline_string.append(line)
                # remove trailing new-line that is part of the \n; delimiter
                multiline_string[-1] = multiline_string[-1].rstrip()
                yield (None, None, "".join(multiline_string), None)
                # Need to process the remainder of the current line
                line = line[1:]
            # Apply regex to the current line
            for match in mmcif_re.finditer(line):
                groups = match.groups()
                if groups != (None, None, None, None):
                    yield groups



##
#
# File: PdbxContainers.py
# Original: 02-Feb-2009 jdw
#
# Update:
# 23-Mar-2011 jdw Added method to rename attributes in category containers.
# 05-Apr-2011 jdw Change cif writer to select double quoting as preferred
# quoting style where possible.
# 16-Jan-2012 jdw Create base class for DataCategory class
# 22-Mar-2012 jdw when append attributes to existing categories update
# existing rows with placeholder null values.
# 2-Sep-2012 jdw add option to avoid embedded quoting that might
# confuse simple parsers.
# 28-Jun-2013 jdw export remove method
# 29-Jun-2013 jdw export remove row method
##
"""A collection of container classes supporting the PDBx/mmCIF storage model.

A base container class is defined which supports common features of
data and definition containers. PDBx data files are organized in
sections called data blocks which are mapped to data containers.
PDBx dictionaries contain definition sections and data sections
which are mapped to definition and data containers respectively.

Data in both PDBx data files and dictionaries are organized in
data categories. In the PDBx syntax individual items or data
identified by labels of the form '_categoryName.attribute_name'.
The terms category and attribute in PDBx jargon are analogous
table and column in relational data model, or class and attribute
in an object oriented data model.

The DataCategory class provides base storage container for instance
data and definition meta data.

"""
import re
from sys import stdout
import traceback


__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"


class CifName:
    """Class of utilities for CIF-style data names."""

    @staticmethod
    def category_part(name) -> str:
        """Get the category part of the name.

        :param str name:  name
        :returns:  category part of name
        """
        tname = ""
        if name.startswith("_"):
            tname = name[1:]
        else:
            tname = name

        i = tname.find(".")
        if i == -1:
            return tname
        else:
            return tname[:i]

    @staticmethod
    def attribute_part(name) -> str:
        """Get the attribute part of the name.

        :param str name:  name
        :returns:  attribute part of name
        """
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1 :]  # noqa : E203


class ContainerBase:
    """Container base class for data and definition objects."""

    def __init__(self, name):
        """Initialize with container name.

        :param str name:  name
        """
        # The enclosing scope of the data container (e.g. data_/save_)
        self.__name = name
        # List of category names within this container -
        self.__object_name_list = []
        # dictionary of DataCategory objects keyed by category name.
        self.__object_catalog = {}
        self.__type = None

    def get_type(self) -> str:
        """Get container type.

        :returns:  container type
        """
        return self.__type

    def set_type(self, type_):
        """Set container type.

        :param str type_:  container type
        """
        self.__type = type_

    @property
    def name(self) -> str:
        """Get container name.

        :returns:  container name
        """
        return self.__name

    def set_name(self, name):
        """Set container name.

        :param str name:  container name
        """
        self.__name = name

    def exists(self, name) -> bool:
        """Determine if object name exists in object catalog.

        :param str name:  object name
        :returns:  whether object exists in object catalog
        """
        return name in self.__object_catalog

    def get_object(self, name):
        """Get object from object catalog.

        :param str name:  object name
        :returns:  object or None
        :rtype:  :class:`~pdbx.containers.DataCategory`
        """
        if name in self.__object_catalog:
            return self.__object_catalog[name]
        else:
            return None

    def get_object_name_list(self) -> list:
        """Get list of object names.

        :returns: list of :class:`~pdbx.containers.DataCategory` objects
        """
        return self.__object_name_list

    def append(self, obj):
        """Add the input object to the current object catalog.
        An existing object of the same name will be overwritten.

        :param obj:  input object to catalog
        :type obj:  :class:`~pdbx.containers.DataCategory`
        """
        if obj.name is not None:
            if obj.name not in self.__object_catalog:
                self.__object_name_list.append(obj.name)
            self.__object_catalog[obj.name] = obj

    def replace(self, obj):
        """Replace an existing object with the input object.

        :param obj:  input object to catalog
        :type obj:  :class:`~pdbx.containers.DataCategory`
        """
        if (obj.name is not None) and (obj.name in self.__object_catalog):
            self.__object_catalog[obj.name] = obj

    def print_it(self, fobj=stdout, type_="brief"):
        """Dump information about container to specified file object.

        :param file fobj:  file object for writing
        :param str type_:  type of summary ("brief" makes it short)
        """
        fobj.write(
            "+ %s container: %30s contains %4d categories\n"
            % (self.get_type(), self.name, len(self.__object_name_list))
        )
        for name in self.__object_name_list:
            fobj.write("--------------------------------------------\n")
            fobj.write("Data category: %s\n" % name)
            if type_ == "brief":
                self.__object_catalog[name].print_it(fobj)
            else:
                self.__object_catalog[name].dump_it(fobj)

    def rename(self, current_name, new_name) -> bool:
        """Change the name of an object in place.

        :param str current_name:  old name for object
        :param str new_name:  new name for object
        :returns:  indicator of whether renaming was successful
        """
        try:
            i = self.__object_name_list.index(current_name)
            self.__object_name_list[i] = new_name
            self.__object_catalog[new_name] = self.__object_catalog[
                current_name
            ]
            self.__object_catalog[new_name].set_name(new_name)
            return True
        except KeyError:
            return False

    def remove(self, current_name) -> bool:
        """Remove object by name.

        :param str current_name:  name of object to remove
        :returns:  True on success or False otherwise.
        """
        try:
            if current_name in self.__object_catalog:
                del self.__object_catalog[current_name]
                i = self.__object_name_list.index(current_name)
                del self.__object_name_list[i]
                return True
            else:
                return False
        except KeyError:
            pass
        return False


class DefinitionContainer(ContainerBase):
    """Container for definitions."""

    def __init__(self, name):
        """Initialize container with name.

        :param str name:  name of container
        """
        super(DefinitionContainer, self).__init__(name)
        self.set_type("definition")

    def is_category(self) -> bool:
        """Determine if container contains category objects.

        :returns:  indicator of whether category objects are in container
        """
        if self.exists("category"):
            return True
        return False

    def is_attribute(self) -> bool:
        """Determine if container contains item objects.

        :returns:  indicator of whether item objects are in container
        """
        if self.exists("item"):
            return True
        return False

    def print_it(self, file_=stdout, type_="brief"):
        """Print information about container to file object.

        :param file file_:  file object for writing
        :param str type_:  type of summary ("brief" makes it short)
        """
        file_.write(
            "Definition container: %30s contains %4d categories\n"
            % (self.name, len(self.get_object_name_list()))
        )
        if self.is_category():
            file_.write("Definition type: category\n")
        elif self.is_attribute():
            file_.write("Definition type: item\n")
        else:
            file_.write("Definition type: undefined\n")

        for name in self.get_object_name_list():
            file_.write("--------------------------------------------\n")
            file_.write("Definition category: %s\n" % name)
            if type_ == "brief":
                self.get_object(name).print_it(file_)
            else:
                self.get_object(name).dumpId(file_)


class DataContainer(ContainerBase):
    """Container class for DataCategory objects."""

    def __init__(self, name):
        """Initialize container with name.

        :param str name:  name of container
        """
        super(DataContainer, self).__init__(name)
        self.set_type("data")
        self.__global_flag = False
        self.__current_row = None

    def invoke_data_block_method(self, method):
        """Invoke a method for the given data block.

        :param str method:  name of method
        """
        self.__current_row = 1
        # TODO - remove exec() commands!
        exec(method.get_inline())

    def set_global(self):
        """Set global flag to True."""
        self.__global_flag = True

    def get_global(self) -> bool:
        """Return global flag."""
        return self.__global_flag


class DataCategoryBase:
    """Base object definition for a data category."""

    def __init__(self, name, attribute_name_list=None, row_list=None):
        """Initialize data category objcet.

        :param str name:  name of data category object
        :param list attribute_name_list:  list of attribute names
        :param list row_list:  list of rows for data category object
        """
        self._name = name
        if row_list is not None:
            self._row_list = row_list
        else:
            self._row_list = []

        if attribute_name_list is not None:
            self._attribute_name_list = attribute_name_list
        else:
            self._attribute_name_list = []
        self._catalog = {}
        self._num_attributes = 0
        self.__setup()

    def __setup(self):
        self._num_attributes = len(self._attribute_name_list)
        self._catalog = {}
        for attribute_name in self._attribute_name_list:
            attribute_name_lower = attribute_name.lower()
            self._catalog[attribute_name_lower] = attribute_name

    def set_row_list(self, row_list):
        """Set row list.

        :param list row_list:  list of rows
        """
        self._row_list = row_list

    def set_attribute_name_list(self, attribute_name_list):
        """Set attribute name list.

        :param list attribute_name_list:  list of attribute names
        """
        self._attribute_name_list = attribute_name_list
        self.__setup()

    def set_name(self, name):
        """Set name.

        :param str name:  object name to set
        """
        self._name = name

    def get(self) -> tuple:
        """Get name, attribute name list, and row list.

        :returns:  tuple of (name, attribute name list, and row list)
        """
        return (self._name, self._attribute_name_list, self._row_list)


class DataCategory(DataCategoryBase):
    """Methods for creating, accessing, formatting PDBx cif data categories."""

    def __init__(self, name, attribute_name_list=None, row_list=None):
        """Initialize object.

        :param str name:  name of object
        :param list attribute_name_list:  list of attribute names
        :param list row_list:  list of rows
        """
        super().__init__(name, attribute_name_list, row_list)
        self.__lfh = stdout
        self.__current_row_index = 0
        self.__current_attribute = None
        self.__avoid_embedded_quoting = False
        self.__whitespace_re = re.compile(r"\s")
        self.__whitespace_quotes_re = re.compile(r"[\s'\"]")
        self.__newline_re = re.compile(r"[\n\r]")
        self.__single_quote_re = re.compile(r"[']")
        self.__whitespace_single_quote_re = re.compile(r"('\s)|(\s')")
        self.__double_quote_re = re.compile(r'["]')
        self.__whitespace_double_quote_re = re.compile(r'("\s)|(\s")')
        self.__integer_re = re.compile(r"^[0-9]+$")
        self.__float_re = re.compile(
            r"^-?(([0-9]+)[.]?|([0-9]*[.][0-9]+))"
            r"([(][0-9]+[)])?([eE][+-]?[0-9]+)?$"
        )
        self.__data_type_list = [
            "DT_NULL_VALUE",
            "DT_INTEGER",
            "DT_FLOAT",
            "DT_UNQUOTED_STRING",
            "DT_ITEM_NAME",
            "DT_DOUBLE_QUOTED_STRING",
            "DT_SINGLE_QUOTED_STRING",
            "DT_MULTI_LINE_STRING",
        ]
        self.__format_type_list = [
            "FT_NULL_VALUE",
            "FT_NUMBER",
            "FT_NUMBER",
            "FT_UNQUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_QUOTED_STRING",
            "FT_MULTI_LINE_STRING",
        ]

    def __getitem__(self, item):
        """Implements list-type functionality.

        Implements op[x] for some special cases:
            item = integer - returns the row in category (normal list behavior)
            item = string - returns the value of attribute 'x' in first row

        :param str item:  item name
        :returns:  object associated with item
        """
        if isinstance(item, int):
            return self._row_list[item]

        elif isinstance(item, str):
            try:
                self.get_attribute_index(item)
                return self._row_list[0][item]
            except (IndexError, KeyError):
                raise KeyError
        raise TypeError(item)

    @property
    def current_attribute(self) -> str:
        """Get current attribute."""
        return self.__current_attribute

    @property
    def current_row_index(self) -> int:
        """Get current row index."""
        return self.__current_row_index

    @property
    def row_list(self) -> list:
        """Get list of rows."""
        return self._row_list

    @property
    def row_count(self) -> int:
        """Get number of rows."""
        return len(self._row_list)

    def get_row(self, index) -> list:
        """Get specified row.

        :param int index:  row index
        :returns: specified row or empty array if row not found.
        """
        try:
            return self._row_list[index]
        except IndexError:
            return []

    def remove_row(self, index) -> bool:
        """Remove specified row.

        :param int index:  index of row to remove
        :returns:  True if successful, False otherwise
        """
        try:
            if index >= 0 and self._row_list:
                del self._row_list[index]
                if self.__current_row_index >= len(self._row_list):
                    self.__current_row_index = len(self._row_list) - 1
                return True
            else:
                pass
        except IndexError:
            pass
        return False

    def get_full_row(self, index) -> list:
        """Return a full row based on the length of the the attribute list.

        :param int index:  index of row to retrieve
        :returns:  row
        """
        try:
            if len(self._row_list[index]) < self._num_attributes:
                self._row_list[index] += ["?"] * (
                    self._num_attributes - len(self._row_list[index])
                )
            return self._row_list[index]
        except IndexError:
            return ["?"] * len(self._num_attributes)

    @property
    def name(self) -> str:
        """Get container name."""
        return self._name

    @property
    def attribute_list(self) -> list:
        """Get list of attributes."""
        return self._attribute_name_list

    @property
    def attribute_count(self) -> int:
        """Get number of attributes."""
        return len(self._attribute_name_list)

    @property
    def attribute_list_with_order(self) -> list:
        """Get list of attributes in order."""
        ordered_list = []
        for i, att in enumerate(self._attribute_name_list):
            ordered_list.append((att, i))
        return ordered_list

    def get_attribute_index(self, attribute_name) -> int:
        """Get index of given attribute.

        :param str attribute_name:  name of attribute
        :returns:  index of attribute
        :raises IndexError:  if attribute not found
        """
        return self._attribute_name_list.index(attribute_name)

    def has_attribute(self, attribute_name) -> bool:
        """Indicate whether container has attribute."""
        return attribute_name in self._attribute_name_list

    @property
    def item_name_list(self) -> list:
        """List of attribute names as fully qualified item names."""
        item_name_list_ = []
        for att in self._attribute_name_list:
            item_name_list_.append("_" + self._name + "." + att)
        return item_name_list_

    def append(self, row):
        """Add row to container.

        :param list row:  row to add
        """
        self._row_list.append(row)

    def append_attribute(self, attribute_name):
        """Add attribute to container.

        :param str attribute_name:  name of attribute to add
        """
        attribute_name_lower = attribute_name.lower()
        if attribute_name_lower in self._catalog:
            index = self._attribute_name_list.index(
                self._catalog[attribute_name_lower]
            )
            self._attribute_name_list[index] = attribute_name
            self._catalog[attribute_name_lower] = attribute_name
        else:
            self._attribute_name_list.append(attribute_name)
            self._catalog[attribute_name_lower] = attribute_name
        self._num_attributes = len(self._attribute_name_list)

    def append_attribute_extend_rows(self, attribute_name):
        """Append attribute and extend rows.

        :param str attribute_name:  name of attribute to add
        """
        attribute_name_lower = attribute_name.lower()
        if attribute_name_lower in self._catalog:
            index = self._attribute_name_list.index(
                self._catalog[attribute_name_lower]
            )
            self._attribute_name_list[index] = attribute_name
            self._catalog[attribute_name_lower] = attribute_name
            self.__lfh.write(
                "Appending existing attribute %s\n" % attribute_name
            )
        else:
            self._attribute_name_list.append(attribute_name)
            self._catalog[attribute_name_lower] = attribute_name
            # add a placeholder to any existing rows for the new attribute.
            for row in self._row_list:
                row.append("?")
        self._num_attributes = len(self._attribute_name_list)

    def get_value(self, attribute_name=None, row_index=None):
        """Get value for specified attribute and row.

        :param str attribute_name:  attribute name
        :param int row_index:  row index
        :returns:  attribute value
        :raises IndexError:  if attribute not found
        """
        if attribute_name is None:
            attribute = self.__current_attribute
        else:
            attribute = attribute_name
        if row_index is None:
            index = self.__current_row_index
        else:
            index = row_index
        if isinstance(attribute, str) and isinstance(index, int):
            return self._row_list[index][
                self._attribute_name_list.index(attribute)
            ]
        raise IndexError(attribute)

    def set_value(self, value, attribute_name=None, row_index=None):
        """Set value of attribute.

        :param value:  value of attribute to set
        :param str attribute_name:  name of attribute
        :param int row_index:  index of row
        """
        if attribute_name is None:
            attribute = self.__current_attribute
        else:
            attribute = attribute_name
        if row_index is None:
            index = self.__current_row_index
        else:
            index = row_index
        if isinstance(attribute, str) and isinstance(index, int):
            try:
                # if row index is out of range - add the rows
                for _ in range(index + 1 - len(self._row_list)):
                    self._row_list.append(self.__empty_row)
                row_len = len(self._row_list[index])
                ind = self._attribute_name_list.index(attribute)
                # extend the list if needed
                if ind >= row_len:
                    self._row_list[index].extend(
                        [None for _ in range(2 * ind - row_len)]
                    )
                self._row_list[index][ind] = value
            except IndexError:
                self.__lfh.write(
                    "DataCategory(setvalue) index error category"
                    " %s attribute %s index %d value %r\n"
                    % (self._name, attribute, index, value)
                )
                traceback.print_exc(file=self.__lfh)
            except ValueError:
                self.__lfh.write(
                    "DataCategory(setvalue) value error category"
                    " %s attribute %s index %d value %r\n"
                    % (self._name, attribute, index, value)
                )
                traceback.print_exc(file=self.__lfh)

    @property
    def __empty_row(self) -> list:
        """Return an empty row."""
        return [None for _ in range(len(self._attribute_name_list))]

    def replace_value(self, old_value, new_value, attribute_name) -> int:
        """Replace the value of the specified attribute.

        :param old_value:  old attribute value
        :param new_value:  new attribute value
        :param str attribute_name:  name of attribute to replace
        :returns:  number of replacements
        """
        num_replace = 0
        if attribute_name not in self._attribute_name_list:
            return num_replace
        ind = self._attribute_name_list.index(attribute_name)
        for row in self._row_list:
            if row[ind] == old_value:
                row[ind] = new_value
                num_replace += 1
        return num_replace

    def replace_substring(self, old_value, new_value, attribute_name) -> bool:
        """Replace substring of value of given attribute.

        :param old_value:  old attribute value
        :param new_value:  new attribute value
        :param str attribute_name:  name of attribute to replace
        :returns:  Boolean flag indicating success.
        """
        replace_ok = False
        if attribute_name not in self._attribute_name_list:
            return replace_ok
        ind = self._attribute_name_list.index(attribute_name)
        for row in self._row_list:
            val = row[ind]
            row[ind] = val.replace(old_value, new_value)
            if val != row[ind]:
                replace_ok = True
        return replace_ok

    def invoke_attribute_method(self, attribute_name, method):
        """Invoke method of current attribute.

        :param str attribute_name:  attribute name
        :param str method:  name of attribute method
        """
        self.__current_row_index = 0
        self.__current_attribute = attribute_name
        self.append_attribute(attribute_name)
        ind = self._attribute_name_list.index(attribute_name)
        if not self._row_list:
            row = [None] * len(self._attribute_name_list) * 2
            row[ind] = None
            self._row_list.append(row)
        for row in self._row_list:
            row_len = len(row)
            if ind >= row_len:
                row.extend([None] * (2 * ind - row_len))
                row[ind] = None
            # TODO - just say "no" to exec()
            exec(method.get_inline())
            self.__current_row_index += 1

    def invoke_category_method(self, method):
        """Invoke method of current category.

        :param str method:  name of method
        """
        self.__current_row_index = 0
        # TODO - remove exec()
        exec(method.get_inline())

    @property
    def max_attribute_list_length(self) -> int:
        """Get maximum attribute list length."""
        max_list = [0] * len(self._attribute_name_list)
        for row in self._row_list:
            for index, value in enumerate(row):
                max_list[index] = max(max_list[index], len(value))
        return max_list

    def rename_attribute(
        self, current_attribute_name, new_attribute_name
    ) -> bool:
        """Change the name of an attribute in place.

        :param str current_attribute_name:  current attribute name
        :param str new_attribute_name:  new attribute name
        :returns:  flag indicating renaming success
        """
        try:
            i = self._attribute_name_list.index(current_attribute_name)
            self._attribute_name_list[i] = new_attribute_name
            del self._catalog[current_attribute_name.lower()]
            self._catalog[new_attribute_name.lower()] = new_attribute_name
            return True
        except KeyError:
            return False

    def print_it(self, file_=stdout):
        """Print container information.

        :param file file_:  file object ready for writing
        """
        file_.write("--------------------------------------------\n")
        file_.write(
            " Category: %s attribute list length: %d\n"
            % (self._name, len(self._attribute_name_list))
        )
        for attr in self._attribute_name_list:
            file_.write(" Category: %s attribute: %s\n" % (self._name, attr))

        file_.write(" Row value list length: %d\n" % len(self._row_list))
        for row in self._row_list[:2]:
            if len(row) == len(self._attribute_name_list):
                for index, value in enumerate(row):
                    file_.write(
                        " %30s: %s ...\n"
                        % (self._attribute_name_list[index], str(value)[:30])
                    )
            else:
                file_.write(
                    "+WARNING - %s data length %d attribute name length %s "
                    "mismatched\n"
                    % (self._name, len(row), len(self._attribute_name_list))
                )

    def dump_it(self, file_=stdout):
        """Dump contents of container.

        :param file file_:  file object ready for writing
        """
        file_.write("--------------------------------------------\n")
        file_.write(
            " Category: %s attribute list length: %d\n"
            % (self._name, len(self._attribute_name_list))
        )
        for attr in self._attribute_name_list:
            file_.write(" Category: %s attribute: %s\n" % (self._name, attr))

        file_.write(" Value list length: %d\n" % len(self._row_list))
        for row in self._row_list:
            for index, value in enumerate(row):
                file_.write(
                    " %30s: %s\n" % (self._attribute_name_list[index], value)
                )

    def __format_pdbx(self, inp) -> str:
        """Format input data following PDBx quoting rules.

        :param str inp:  input data
        :returns:  formatted data
        """
        try:
            if inp is None:
                return ("?", "DT_NULL_VALUE")
            # pure numerical values are returned as unquoted strings
            if isinstance(inp, int) or self.__integer_re.search(str(inp)):
                return ([str(inp)], "DT_INTEGER")
            if isinstance(inp, float) or self.__float_re.search(str(inp)):
                return ([str(inp)], "DT_FLOAT")
            # null value handling
            if inp in (".", "?"):
                return (
                    self.__double_quoted_list(inp),
                    "DT_DOUBLE_QUOTED_STRING",
                )
            if inp == "":
                return (["."], "DT_NULL_VALUE")
            # Contains white space or quotes ?
            if not self.__whitespace_quotes_re.search(inp):
                if inp.startswith("_"):
                    return (self.__double_quoted_list(inp), "DT_ITEM_NAME")
                else:
                    return ([str(inp)], "DT_UNQUOTED_STRING")
            else:
                if self.__newline_re.search(inp):
                    return (
                        self.__semicolon_quoted_list(inp),
                        "DT_MULTI_LINE_STRING",
                    )
                else:
                    if self.__avoid_embedded_quoting:
                        # change priority to choose double quoting where
                        # possible.
                        if not self.__double_quote_re.search(
                            inp
                        ) and not self.__whitespace_single_quote_re.search(
                            inp
                        ):
                            return (
                                self.__double_quoted_list(inp),
                                "DT_DOUBLE_QUOTED_STRING",
                            )
                        elif not self.__single_quote_re.search(
                            inp
                        ) and not self.__whitespace_double_quote_re.search(
                            inp
                        ):
                            return (
                                self.__single_quoted_list(inp),
                                "DT_SINGLE_QUOTED_STRING",
                            )
                        else:
                            return (
                                self.__semicolon_quoted_list(inp),
                                "DT_MULTI_LINE_STRING",
                            )
                    else:
                        # change priority to choose double quoting where
                        # possible.
                        if not self.__double_quote_re.search(inp):
                            return (
                                self.__double_quoted_list(inp),
                                "DT_DOUBLE_QUOTED_STRING",
                            )
                        elif not self.__single_quote_re.search(inp):
                            return (
                                self.__single_quoted_list(inp),
                                "DT_SINGLE_QUOTED_STRING",
                            )
                        else:
                            return (
                                self.__semicolon_quoted_list(inp),
                                "DT_MULTI_LINE_STRING",
                            )
        except ValueError:
            traceback.print_exc(file=self.__lfh)

    def __data_type_pdbx(self, inp) -> str:
        """Detect the PDBx data type.

        :param str inp:  input data
        :returns:  data type
        """
        if inp is None:
            return "DT_NULL_VALUE"
        # pure numerical values are returned as unquoted strings
        if isinstance(inp, int) or self.__integer_re.search(str(inp)):
            return "DT_INTEGER"
        if isinstance(inp, float) or self.__float_re.search(str(inp)):
            return "DT_FLOAT"
        # null value handling
        if inp in (".", "?"):
            return "DT_DOUBLE_QUOTED_STRING"
        if inp == "":
            return "DT_NULL_VALUE"
        # Contains white space or quotes ?
        if not self.__whitespace_quotes_re.search(inp):
            if inp.startswith("_"):
                return "DT_ITEM_NAME"
            else:
                return "DT_UNQUOTED_STRING"
        else:
            if self.__newline_re.search(inp):
                return "DT_MULTI_LINE_STRING"
            else:
                if self.__avoid_embedded_quoting:
                    if not self.__single_quote_re.search(
                        inp
                    ) and not self.__whitespace_double_quote_re.search(inp):
                        return "DT_DOUBLE_QUOTED_STRING"
                    elif not self.__double_quote_re.search(
                        inp
                    ) and not self.__whitespace_single_quote_re.search(inp):
                        return "DT_SINGLE_QUOTED_STRING"
                    else:
                        return "DT_MULTI_LINE_STRING"
                else:
                    if not self.__single_quote_re.search(inp):
                        return "DT_DOUBLE_QUOTED_STRING"
                    elif not self.__double_quote_re.search(inp):
                        return "DT_SINGLE_QUOTED_STRING"
                    else:
                        return "DT_MULTI_LINE_STRING"

    @staticmethod
    def __single_quoted_list(inp) -> str:
        """Generate a single-quoted list from the input."""
        return ["'"] + [inp] + ["'"]

    @staticmethod
    def __double_quoted_list(inp) -> str:
        """Generate a double-quoted list from the input."""
        return ['"'] + [inp] + ['"']

    @staticmethod
    def __semicolon_quoted_list(inp) -> str:
        """Generate a semicolon-delimited quoted list from the input."""
        if inp[-1] == "\n":
            return ["\n", ";"] + [inp] + [";", "\n"]
        else:
            return ["\n", ";"] + [inp] + ["\n", ";", "\n"]

    def get_value_formatted(self, attribute_name=None, row_index=None) -> str:
        """Get formatted version of value.

        :param str attribute_name:  attribute name
        :param int row_index:  row index
        :returns:  formatted value
        """
        if attribute_name is None:
            attribute = self.__current_attribute
        else:
            attribute = attribute_name
        if row_index is None:
            index = self.__current_row_index
        else:
            index = row_index
        if isinstance(attribute, str) and isinstance(index, int):
            try:
                list_, _ = self.__format_pdbx(
                    self._row_list[index][
                        self._attribute_name_list.index(attribute)
                    ]
                )
                return "".join(list_)
            except IndexError:
                self.__lfh.write(
                    "attribute_name %s index %r rowdata %r\n"
                    % (attribute_name, index, self._row_list[index])
                )
                raise IndexError
        raise TypeError(attribute)

    def get_value_formatted_by_index(self, attribute_index, row_index) -> str:
        """Get value formatted by index.

        :param str attribute_name:  attribute name
        :param int row_index:  row index
        :returns:  formatted value
        """
        list_, _ = self.__format_pdbx(
            self._row_list[row_index][attribute_index]
        )
        return "".join(list_)

    def get_max_attribute_list_length(self, steps=1) -> int:
        """Get maximum length of attribute value list.

        :param int steps:  step size for iterating through rows
        :returns:  attribute value list max length
        """
        max_list = [0] * len(self._attribute_name_list)
        for _ in self._row_list[::steps]:
            for index, value in enumerate(self._attribute_name_list):
                max_list[index] = max(max_list[index], len(str(value)))
        return max_list

    def get_format_type_list(self, steps=1) -> str:
        """Get a formatted type list.

        :param int  steps:  step size for iterating through rows
        :returns:  formatted type list
        """
        try:
            current_data_type_list = ["DT_NULL_VALUE"] * len(
                self._attribute_name_list
            )
            for row in self._row_list[::steps]:
                for index, value in enumerate(self._attribute_name_list):
                    data_type = self.__data_type_pdbx(value)
                    data_index = self.__data_type_list.index(data_type)
                    current_type = current_data_type_list[index]
                    current_index = self.__data_type_list.index(current_type)
                    current_index = max(current_index, data_index)
                    current_data_type_list[index] = self.__data_type_list[
                        current_index
                    ]
            # Map the format types to the data types
            current_format_type_list = []
            for data_type in current_data_type_list:
                index = self.__data_type_list.index(data_type)
                current_format_type_list.append(self.__format_type_list[index])
        except IndexError:
            self.__lfh.write(
                "PdbxDataCategory(get_format_type_list) ++Index error at "
                "index %d in row %r\n" % (index, row)
            )
        return current_format_type_list, current_data_type_list

    @property
    def get_format_type_list_x(self) -> str:
        """Alternate version of format type list."""
        current_data_type_list = ["DT_NULL_VALUE"] * len(
            self._attribute_name_list
        )
        for _ in self._row_list:
            for index, value in enumerate(self._attribute_name_list):
                data_type = self.__data_type_pdbx(value)
                data_index = self.__data_type_list.index(data_type)
                current_type = current_data_type_list[index]
                current_index = self.__data_type_list.index(current_type)
                current_index = max(current_index, data_index)
                current_data_type_list[index] = self.__data_type_list[
                    current_index
                ]
        # Map the format types to the data types
        current_format_type_list = []
        for data_type in current_data_type_list:
            index = self.__data_type_list.index(data_type)
            current_format_type_list.append(self.__format_type_list[index])
        return current_format_type_list, current_data_type_list



"""PDBX Version number.

Store the version here so:

* we don't load dependencies by storing it in :file:`__init__.py`
* we can import it in setup.py for the same reason
* we can import it into your module

"""
__version__ = "2.0.1"



"""Python packaging for PDBx/mmCIF utilities."""
import setuptools

# NOTE: The following reads the version number and makes
#       if available to the packaging tools before installation.
#       REF: https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package  # noqa: E501
with open("pdbx/_version.py") as fobj:
    exec(fobj.read())


with open("README.md", "r") as fobj:
    LONG_DESCRIPTION = fobj.read()


setuptools.setup(
    name="mmcif-pdbx",
    version=__version__,  # noqa: F821
    author="John Westbrook",
    author_email="jwest@rcsb.rutgers.edu",
    maintainer="Nathan Baker",
    maintainer_email="nathanandrewbaker@gmail.com",
    description="Python utilities for PDBx/mmCIF storage model",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/Electrostatics/mmcif_pdbx",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
    python_requires=">=3.6",
    tests_require=["pytest"],
    test_suite="tests",
    zip_safe=True,
)
